from starlette.endpoints import HTTPEndpoint
from starlette.responses import HTMLResponse, Response, RedirectResponse, JSONResponse
from starlette.requests import Request
from starlette.responses import Response, RedirectResponse

import otree.bots.browser as browser_bots
import otree.channels.utils as channel_utils
import otree.common
import otree.constants
import otree.models
import otree.views.admin
import otree.views.cbv
import otree.views.mturk
from otree.common import (
    make_hash,
    BotError,
    GlobalState,
)
from otree.database import NoResultFound, db, dbq
from otree.i18n import core_gettext as _
from otree.models import Participant, Session
from otree.models_concrete import ParticipantVarsFromREST
from otree.mturk_client import TurkClient
from otree.room import ROOM_DICT
from otree.templating import render
from otree.views.abstract import GenericWaitPageMixin


def no_participants_left_http_response():
    '''
    this function exists because i'm not sure if Http response objects can be reused
    better to return 404 so browser bot client & tests can recognize it
    '''
    # Translators: for example this is shown if you create a session for 10
    # participants. The 11th person to click will get this message
    # It means there is no space for you.
    return Response(_("Session is full."), status_code=404)


class OutOfRangeNotification(HTTPEndpoint):
    url_pattern = '/OutOfRangeNotification/{code}'

    def get(self, request):
        code = request.path_params['code']
        participant = db.get_or_404(Participant, code=code)
        if participant.is_browser_bot:
            session = participant.session
            has_next_submission = browser_bots.enqueue_next_post_data(
                participant_code=participant.code
            )

            if has_next_submission:
                msg = (
                    'Finished the last page, '
                    'but the bot is still trying '
                    'to submit more pages.'
                )
                raise BotError(msg)

            browser_bots.send_completion_message(
                session_code=session.code, participant_code=code
            )

        return render('otree/OutOfRangeNotification.html', {})


class GBATInactive(HTTPEndpoint):
    url_pattern = '/GBATInactive'

    def get(self, request):
        return render('otree/GBATInactive.html', {})


class InitializeParticipant(HTTPEndpoint):

    url_pattern = '/InitializeParticipant/{code}'

    def get(self, request: Request):
        """anything essential should be done in"""
        code = request.path_params['code']
        pp = db.get_or_404(Participant, code=code)
        query_params = request.query_params
        label = query_params.get(otree.constants.participant_label)
        pp.initialize(label)

        first_url = pp._url_i_should_be_on()
        return RedirectResponse(first_url)


class MTurkStart(HTTPEndpoint):

    url_pattern = r"/MTurkStart/{code}"

    def get(self, request: Request):
        code = request.path_params['code']
        session = self.session = db.get_or_404(Session, code=code)
        GET = request.query_params
        try:
            assignment_id = GET['assignmentId']
            worker_id = GET['workerId']
        except KeyError:
            return Response(
                'URL is missing assignmentId or workerId parameter', status_code=404
            )
        qual_id = session.config['mturk_hit_settings'].get('grant_qualification_id')
        use_sandbox = session.mturk_use_sandbox
        if qual_id and not use_sandbox:
            # this is necessary because MTurk's qualification requirements
            # don't prevent 100% of duplicate participation. See:
            # https://groups.google.com/forum/#!topic/otree/B66HhbFE9ck

            previous_participation = (
                dbq(Participant)
                .join(Session)
                .filter(
                    Participant.session != session,
                    Session.mturk_qual_id == qual_id,
                    Participant.mturk_worker_id == worker_id,
                )
                .scalar()
                is not None
            )
            if previous_participation:
                return Response('You have already accepted a related HIT')

            # if using sandbox, there is no point in granting quals.
            # https://groups.google.com/forum/#!topic/otree/aAmqTUF-b60

            # don't pass request arg, because we don't want to show a message.
            # using the fully qualified name because that seems to make mock.patch work

            # seems OK to assign this multiple times
            TurkClient.assign_qualification(
                dict(
                    QualificationTypeId=qual_id,
                    WorkerId=worker_id,
                    # Mturk complains if I omit IntegerValue
                    IntegerValue=1,
                ),
                use_sandbox=use_sandbox,
            )

        try:
            # just check if this worker already game, but
            # don't filter for assignment, because maybe they already started
            # and returned the previous assignment
            # in this case, we should assign back to the same participant
            # so that we don't get duplicates in the DB, and so people
            # can't snoop and try the HIT first, then re-try to get a bigger bonus
            pp = self.session.pp_set.filter_by(mturk_worker_id=worker_id).one()
        except NoResultFound:
            pp = self.session.pp_set.filter_by(visited=False).order_by('id').first()
            if not pp:
                return no_participants_left_http_response()

            # 2014-10-17: needs to be here even if it's also set in
            # the next view to prevent race conditions
            # this needs to be inside the lock
            pp.visited = True
            pp.mturk_worker_id = worker_id
        # reassign assignment_id, even if they are returning, because maybe they accepted
        # and then returned, then re-accepted with a different assignment ID
        # if it's their second time
        pp.mturk_assignment_id = assignment_id
        return _redirect_to_start_url(pp)


def _redirect_to_start_url(participant):
    start_url = participant._start_url()
    return RedirectResponse(start_url, status_code=302)


def get_participant_by_label(session, label):
    q = session.pp_set
    if label:
        try:
            return q.filter_by(label=label).one()
        except NoResultFound:
            pass
    return q.filter_by(visited=False).order_by('id').first()


def get_participant_by_cookies(session, cookies):
    cookie_name = 'session_{}_participant'.format(session.code)
    code = cookies.get(cookie_name)
    # this could return None
    if code:
        return Participant.objects_filter(code=code).first()
    participant = session.pp_set.filter_by(visited=False).order_by('id').first()
    if participant:
        cookies[cookie_name] = participant.code
        return participant


def mark_visited_and_record_label(participant, label):
    if not participant:
        return

    # needs to be here even if it's also set in
    # the next view to prevent race conditions
    participant.visited = True
    if label:
        participant.set_label(label)

    return participant


def participant_or_none_if_exceeded(session, *, label, cookies=None):
    '''pass request.session as an arg if you want to get/set a cookie'''
    if cookies is None:
        participant = get_participant_by_label(session, label)
    else:
        participant = get_participant_by_cookies(session, cookies)
    if not participant:
        return

    return mark_visited_and_record_label(participant, label)


class SessionWideLink(HTTPEndpoint):
    """
    Better to have just 1 endpoint.
    In demo mode, we don't use cookies.
    With a real session, we do use cookies.
    We never need to have both.
    And don't have to think about what if the participant changes the URL.
    """

    url_pattern = '/join/{anonymous_code}'

    use_cookies: bool

    def get_participant(self, request, label):
        raise NotImplementedError

    def get(self, request: Request):
        anonymous_code = request.path_params['anonymous_code']
        session = db.get_or_404(Session, _anonymous_code=anonymous_code)
        label = request.query_params.get('participant_label')

        # is this priority right? maybe cookies should override label?
        # but maybe a participant first clicks the wrong link,
        # they shouldn't be stuck and forbidden from participating.

        if label:
            participant = get_participant_by_label(session, label)
        elif not session.is_demo:
            participant = get_participant_by_cookies(session, request.session)
        else:
            participant = get_participant_by_label(session, None)

        if not participant:
            return no_participants_left_http_response()
        mark_visited_and_record_label(participant, label)

        return _redirect_to_start_url(participant)


def validate_room_visitor(room, query_params: dict) -> dict:
    errors = {}
    if room.has_participant_labels:
        label = query_params.get('participant_label', '')
        if not label:
            errors['participant_label'] = 'Please enter a participant label'
        elif label not in room.get_participant_labels():
            errors['participant_label'] = 'This participant label was not found'
        elif room.use_secure_urls:
            hash = query_params.get('hash')
            if not hash or hash != make_hash(label):
                errors['hash'] = (
                    'Invalid hash parameter. use_secure_urls is True, '
                    'so you must use the participant-specific URL.'
                )
    # later, validate custom params here.
    return errors


class AssignVisitorToRoom(GenericWaitPageMixin, HTTPEndpoint):

    url_pattern = '/room/{room_name}'

    def get(self, request: Request):
        room_name = request.path_params['room_name']
        try:
            room = ROOM_DICT[room_name]
        except KeyError:
            return Response('Invalid room specified in url', status_code=404)

        if not request.query_params.get('welcome_page_ok'):
            return render(
                room.welcome_page or "otree/RoomWelcomePage.html",
                dict(has_participant_label_file=room.has_participant_labels),
            )

        errors = validate_room_visitor(room, request.query_params)

        if errors:
            return HTMLResponse(f"<pre>{errors}</pre>", status_code=404)

        # labels are still allowed even if the room has no participant_labels_file
        label = request.query_params.get('participant_label', '')

        session = room.get_session()

        if session is None:
            self.tab_unique_id = otree.common.random_chars_join_code()
            self._socket_url = channel_utils.room_participant_path(
                room_name=room_name,
                participant_label=label,
                # random chars in case the participant has multiple tabs open
                tab_unique_id=self.tab_unique_id,
            )
            return render(
                "otree/WaitPageGeneric.html",
                dict(
                    view=self,
                    title_text=_('Please wait'),
                    body_text=_('Waiting for your session to begin'),
                ),
            )

        if label:
            participant = get_participant_by_label(session, label)
        else:
            participant = get_participant_by_cookies(session, request.session)

        if not participant:
            return no_participants_left_http_response()

        mark_visited_and_record_label(participant, label)
        if label:  # whether the room has participant labels or not
            passed_vars = ParticipantVarsFromREST.objects_filter(
                room_name=room_name, participant_label=label
            ).first()
            if passed_vars:
                participant.vars.update(passed_vars.vars)
                db.delete(passed_vars)

        return _redirect_to_start_url(participant)

    def socket_url(self):
        return self._socket_url

    async def post(self, request: Request):
        room_name = request.path_params['room_name']
        try:
            room = ROOM_DICT[room_name]
        except KeyError:
            return Response('Invalid room specified in url', status_code=404)

        # Parse JSON data from the request
        try:
            json_data = await request.json()
        except:
            return Response('Invalid JSON data', status_code=400)

        # Merge query params with JSON data for validation
        query_params = dict(request.query_params)
        query_params.update(json_data)

        # Validate the data
        errors = validate_room_visitor(room, query_params)
        if errors:
            return JSONResponse({'errors': errors})

        # If validation passes, return success
        return JSONResponse({'status': 'ok'})


class BrowserBotStartLink(GenericWaitPageMixin, HTTPEndpoint):
    '''should i move this to another module?
    because the rest of these views are accessible without password login.
    '''

    # remote CLI browser bots won't work if this takes an admin_secret_code param because
    # SECRET_KEY might be different on the server.
    url_pattern = '/browser_bot_start'

    def get(self, request):

        session_code = GlobalState.browser_bots_launcher_session_code
        if session_code:
            try:
                session = Session.objects_get(code=session_code)
            except NoResultFound:
                # maybe it's an old session
                pass
            else:
                participant = (
                    session.pp_set.filter_by(visited=False).order_by('id').first()
                )
                if not participant:
                    return no_participants_left_http_response()

                # 2014-10-17: needs to be here even if it's also set in
                # the next view to prevent race conditions
                participant.visited = True
                return _redirect_to_start_url(participant)
        ctx = dict(
            view=self,
            title_text='Please wait',
            body_text='Waiting for browser bots session to begin',
        )
        return render("otree/WaitPageGeneric.html", ctx)

    def socket_url(self):
        return '/browser_bot_wait/'



import logging

from .base import BaseCommand

from otree.session import create_session
from otree.room import ROOM_DICT


logger = logging.getLogger('otree')


class Command(BaseCommand):
    help = "oTree: Create a session."

    def add_arguments(self, parser):
        parser.add_argument('session_config_name', help="The session config name")
        parser.add_argument(
            'num_participants',
            type=int,
            help="Number of participants for the created session",
        )
        parser.add_argument(
            "--room",
            action="store",
            dest="room_name",
            default=None,
            help="Name of room to create the session in",
        )

    def handle(self, session_config_name, num_participants, room_name, **kwargs):

        session = create_session(
            session_config_name=session_config_name, num_participants=num_participants,
        )

        if room_name:
            room = ROOM_DICT[room_name]
            room.set_session(session)
            logger.info(
                "Created session with code {} in room '{}'\n".format(
                    session.code, room_name
                )
            )
        else:
            logger.info("Created session with code {}\n".format(session.code))



from decimal import Decimal, ROUND_HALF_UP
from otree import settings
from otree.i18n import format_number, convert_decimal_separator


def _to_decimal(amount):
    if isinstance(amount, Decimal):
        return amount
    elif isinstance(amount, float):
        return Decimal.from_float(amount)
    else:
        return Decimal(amount)


class DecimalValue(Decimal):
    """
    Generalized decimal type with config-based precision and formatting.
    Completely separate from Currency for backward compatibility.

    Usage:
        score = DecimalValue(0.856, config='percent')
        hours = dec(2.5, config='hours')  # Using convenience function
    """

    def __new__(cls, amount, config):
        if amount is None:
            raise ValueError('Cannot convert None to decimal')
        if config not in settings.DECIMAL_CONFIGS_DICT:
            raise ValueError(f'Unknown decimal config: {config}')
        instance = Decimal.__new__(cls, cls._sanitize(amount, config))
        instance._config = config
        return instance

    @classmethod
    def _sanitize(cls, amount, config):
        if isinstance(amount, DecimalValue) and amount._config == config:
            return amount
        cfg = settings.DECIMAL_CONFIGS_DICT[config]
        try:
            places = cfg['places']
        except KeyError:
            msg = f"Config '{config}' is missing required key 'places'"
            raise ValueError(msg)
        quant = Decimal('0.1') ** places
        return _to_decimal(amount).quantize(quant, rounding=ROUND_HALF_UP)

    # Support for pickling
    def __reduce__(self):
        return (self.__class__, (Decimal.__str__(self), self._config))

    # Immutable
    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def __float__(self):
        return float(Decimal(self))

    def __str__(self):
        return self._format_for_display()

    def __repr__(self):
        return f'{Decimal(self).normalize()}{self._config}'

    def __eq__(self, other):
        if isinstance(other, DecimalValue):
            # Only equal if same config and same value
            return self._config == other._config and Decimal.__eq__(self, other)
        elif isinstance(other, (int, float, Decimal)):
            return Decimal.__eq__(self, self._sanitize(other, self._config))
        else:
            return False

    __hash__ = Decimal.__hash__

    def _format_with_places(self, min_places, max_places):
        """Helper to format with specific decimal places and apply custom function"""
        # Round to max_places
        quant = Decimal('0.1') ** max_places
        rounded = Decimal(self).quantize(quant, rounding=ROUND_HALF_UP)

        # Format with Python's thousand separator (comma) and max_places decimals
        fmt = "{:,.%df}" % max_places
        result = fmt.format(rounded)

        # Strip trailing zeros down to min_places (before converting separators)
        if min_places < max_places and '.' in result:
            parts = result.split('.')
            decimal_part = parts[1].rstrip('0')
            # Ensure we keep at least min_places
            decimal_part = decimal_part.ljust(min_places, '0')
            if decimal_part:
                result = parts[0] + '.' + decimal_part
            else:
                result = parts[0]

        # Convert ',' to THOUSAND_SEPARATOR and '.' to DECIMAL_SEPARATOR
        result = convert_decimal_separator(result)

        # Apply custom function if defined
        cfg = settings.DECIMAL_CONFIGS_DICT[self._config]
        try:
            display_cfg = cfg['display']
        except KeyError:
            msg = f"Config '{self._config}' is missing required key 'display'"
            raise ValueError(msg)
        custom_func = display_cfg.get('function')
        if custom_func:
            result = custom_func(result, self)

        return result

    def _prepare_operand(self, other):
        """Convert operand to Decimal for arithmetic"""
        if isinstance(other, Decimal):
            return other
        return _to_decimal(other)

    @staticmethod
    def _make_binary_operator(name):
        """Factory for creating arithmetic operator methods"""
        method = getattr(Decimal, name, None)

        def binary_function(self, other):
            other = self._prepare_operand(other)
            result = method(self, other)
            return self._make_result(result, other)

        return binary_function

    def _format_for_display(self):
        """Format using config's display settings"""
        cfg = settings.DECIMAL_CONFIGS_DICT[self._config]
        try:
            display_config = cfg['display']
        except KeyError:
            msg = f"Config '{self._config}' is missing required key 'display'"
            raise ValueError(msg)
        try:
            min_places = display_config['min_places']
            max_places = display_config['max_places']
        except KeyError as e:
            msg = f"Config '{self._config}' display is missing required key {e}"
            raise ValueError(msg)
        return self._format_with_places(min_places, max_places)

    def _make_result(self, result, other):
        """DRY helper for arithmetic operations"""
        if not isinstance(other, DecimalValue):
            # Operating with scalar preserves config
            return DecimalValue(result, config=self._config)
        if self._config == other._config:
            # Same config preserves config
            return DecimalValue(result, config=self._config)
        # Different configs return plain Decimal
        return Decimal(result)

    # Arithmetic operations - generated by _make_binary_operator
    __add__ = _make_binary_operator('__add__')
    __radd__ = _make_binary_operator('__radd__')
    __sub__ = _make_binary_operator('__sub__')
    __rsub__ = _make_binary_operator('__rsub__')
    __mul__ = _make_binary_operator('__mul__')
    __rmul__ = _make_binary_operator('__rmul__')
    __truediv__ = _make_binary_operator('__truediv__')
    __rtruediv__ = _make_binary_operator('__rtruediv__')
    __floordiv__ = _make_binary_operator('__floordiv__')
    __rfloordiv__ = _make_binary_operator('__rfloordiv__')
    __mod__ = _make_binary_operator('__mod__')
    __rmod__ = _make_binary_operator('__rmod__')
    __pow__ = _make_binary_operator('__pow__')
    __rpow__ = _make_binary_operator('__rpow__')

    def __neg__(self):
        return DecimalValue(Decimal.__neg__(self), config=self._config)

    def __pos__(self):
        return DecimalValue(Decimal.__pos__(self), config=self._config)

    def __abs__(self):
        return DecimalValue(Decimal.__abs__(self), config=self._config)


def dec(amount, config):
    """Convenience constructor for DecimalValue"""
    return DecimalValue(amount, config=config)



# note-to-self
# like common, but can import models
import importlib.util
import json
import os
import time
from dataclasses import dataclass, asdict
from pathlib import Path

from starlette.staticfiles import StaticFiles

from otree import settings
from otree.database import db
from otree.models_concrete import PageTimeBatch
from otree.currency import Currency, RealWorldCurrency
from otree.decimal import DecimalValue

import typing


@dataclass
class TimeSpentRow:
    session_code: str
    participant_id_in_session: int
    participant_code: str
    page_index: int
    app_name: str
    page_name: str
    epoch_time_completed: int
    round_number: int
    timeout_happened: int
    is_wait_page: int


page_completion_buffer = []
page_completion_last_write = 0

BUFFER_SIZE = 50


def write_row_to_page_buffer(row: TimeSpentRow):
    d = asdict(row)
    row = ','.join(map(str, d.values())) + '\n'

    page_completion_buffer.append(row)
    if (
        len(page_completion_buffer) > BUFFER_SIZE
        or time.time() - page_completion_last_write > 60 * 2
    ):
        write_page_completion_buffer()


def make_page_completion_row(
    *,
    view,
    app_name,
    participant__id_in_session,
    participant__code,
    session_code,
    is_wait_page,
):
    now = int(time.time())
    row = TimeSpentRow(
        app_name=app_name,
        page_index=view._index_in_pages,
        page_name=type(view).__name__,
        epoch_time_completed=now,
        round_number=view.round_number,
        participant_id_in_session=participant__id_in_session,
        participant_code=participant__code,
        session_code=session_code,
        timeout_happened=int(bool(getattr(view, 'timeout_happened', False))),
        is_wait_page=is_wait_page,
    )
    write_row_to_page_buffer(row)


def write_page_completion_buffer():
    global page_completion_last_write
    db.add(PageTimeBatch(text=''.join(page_completion_buffer)))
    page_completion_last_write = time.time()
    page_completion_buffer.clear()


class OTreeStaticFiles(StaticFiles):
    # copied from starlette, just to change 'statics' to 'static',
    # and to fail silently if the dir does not exist.
    def get_directories(self, directory, packages):
        directories = []
        if directory is not None:
            directories.append(directory)

        for package in packages or []:
            spec = importlib.util.find_spec(package)
            assert (
                spec is not None and spec.origin is not None
            ), f"Package {package!r} could not be found, or maybe __init__.py is missing"
            package_directory = os.path.normpath(
                os.path.join(spec.origin, "..", "static")
            )
            if os.path.isdir(package_directory):
                directories.append(package_directory)

        return directories

    def assert_file_exists(self, path):
        if path in existing_filenames_cache:
            return
        for _dir in self.all_directories:
            if Path(_dir, path).is_file():
                existing_filenames_cache.add(path)
                return
        raise FileNotFoundError(path)

    def urls_of_static_files(self, dirpath, extension):
        from otree.asgi import app

        if extension.startswith('.'):
            raise Exception("Extension should not contain a dot")

        for root in self.all_directories:
            _dir = Path(root, dirpath)
            if _dir.is_dir():
                for _file in _dir.glob(f'*.{extension}'):
                    relpath = _file.relative_to(root)
                    yield app.router.url_path_for('static', path=relpath)


existing_filenames_cache = set()


def url_of_static_file(path):
    """
    - better than hardcoding '/static/', which will fail silently if the file
    doesn't exist.
    - there is the {{ static }} tag in templates.
      good to have this in python code for consistency,
      rather than making people construct it manually.
    - useful for:
        - Trials, which are generated on the server side.
          this would be useful for 2 situations:
        - live pages, where {% static %} can't be used because the template was already rendered
        - js_vars (don't want {% static %} mixed in with JS code)

    naming:
    - it shouldn't start with
    'static' because that would distract from @staticmethod in autocomplete,
    which is much more important.
    - url_of_static is more specific than url_for_static (which looks like vars_for_template but works differently)

    """
    from otree.asgi import app

    static_files_app.assert_file_exists(path)
    return app.router.url_path_for('static', path=path)


def urls_of_static_files(dir, extension):
    """
    e.g.:
    >>> urls_of_static_files('emojis', 'png')
    ['/static/emojis/1.png', '/static/emojis/2.png']

    I don't know how useful this would be because usually people have (a) some metadata about the file,
    such as what it contains. also for DB storage & record-keeping,
    people would want to store the filename, e.g. '1.png',
    rather than the full URL path, e.g. '/static/emojis/1.png'

    But if we see a lot of people just trying to include all images in a dir,
    then we can add this to API.
    """
    return static_files_app.urls_of_static_files(dir, extension)


static_files_app = OTreeStaticFiles(
    directory='_static', packages=['otree'] + settings.OTREE_APPS
)


class OTreeJSONEncoder(json.JSONEncoder):
    """JSON encoder that handles Currency and DecimalValue types"""

    def default(self, obj):

        if isinstance(obj, (Currency, RealWorldCurrency)):
            if obj.get_num_decimal_places() == 0:
                return int(obj)
            return float(obj)
        if isinstance(obj, DecimalValue):
            return float(obj)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def json_dumps(obj):
    """JSON dumps with oTree type support"""
    return json.dumps(obj, cls=OTreeJSONEncoder)


safe_json = json_dumps



from sqlalchemy.ext.declarative import declared_attr
import copy
import time
from collections import defaultdict

from sqlalchemy import Column as C, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.sql import sqltypes as st
from sqlalchemy.sql.functions import func


import otree.common
import otree.database
from otree.constants import BaseConstants
from otree.common import (
    get_models_module,
    in_round,
    in_rounds,
    get_constants,
    get_builtin_constant,
)
from otree.common import has_group_by_arrival_time
from otree.database import db, dbq, values_flat, SPGModel, MixinSessionFK


class GroupMatrixError(ValueError):
    pass


class RoundMismatchError(GroupMatrixError):
    pass


class BaseSubsession(SPGModel, MixinSessionFK):
    __abstract__ = True

    round_number = C(
        st.Integer,
        index=True,
    )

    def in_round(self, round_number):
        return in_round(type(self), round_number, session=self.session)

    def in_rounds(self, first, last):
        return in_rounds(type(self), first, last, session=self.session)

    def in_previous_rounds(self):
        return self.in_rounds(1, self.round_number - 1)

    def in_all_rounds(self):
        return self.in_previous_rounds() + [self]

    def get_groups(self):
        return list(self.group_set.order_by('id_in_subsession'))

    def get_players(self):
        return list(self.player_set.order_by('id'))

    def _get_group_matrix(self, objects):
        Player = self._PlayerClass()
        Group = self._GroupClass()
        players = (
            dbq(Player)
            .join(Group)
            .filter(Player.subsession == self)
            .order_by(Group.id_in_subsession, 'id_in_group')
        )
        d = defaultdict(list)
        for p in players:
            d[p.group.id_in_subsession].append(p if objects else p.id_in_subsession)
        return list(d.values())

    def get_group_matrix(self, objects=False):
        return self._get_group_matrix(objects=objects)

    def set_group_matrix(self, matrix):
        """
        warning: this deletes the groups and any data stored on them
        """

        try:
            sample_item = matrix[0][0]
        except TypeError:
            raise GroupMatrixError('Group matrix must be a list of lists.') from None

        if isinstance(sample_item, SPGModel):
            matrix = [[p.id_in_subsession for p in row] for row in matrix]

        ids_flat = [iis for row in matrix for iis in row]

        ids_flat = sorted(ids_flat)
        players_from_db = self.get_players()

        if not ids_flat == list(range(1, len(players_from_db) + 1)):
            msg = 'The matrix of integers either has duplicate or missing elements.'
            raise GroupMatrixError(msg) from None

        matrix = [[players_from_db[iis - 1] for iis in row] for row in matrix]

        self.player_set.update({self._PlayerClass().group_id: None})
        self.group_set.delete()

        GroupClass = self._GroupClass()
        for i, row in enumerate(matrix, start=1):
            group = GroupClass.objects_create(
                subsession=self,
                id_in_subsession=i,
                session=self.session,
                round_number=self.round_number,
            )
            # this line causes
            # SAWarning: Identity map already had an identity for (<class 'set_group_matrix2.Group'>, (10,), None), replacing it with newly flushed object.   Are there load operations occurring inside of an event handler within the flush?
            #   "within the flush?" % (instance_key,)
            group.set_players(row)

    def group_like_round(self, round_number):
        previous_round: BaseSubsession = self.in_round(round_number)
        group_matrix = previous_round.get_group_matrix()
        self.set_group_matrix(group_matrix)

    @property
    def _Constants(self) -> BaseConstants:
        return get_constants(self.get_folder_name())

    def _GroupClass(self):
        return get_models_module(self.get_folder_name()).Group

    def _PlayerClass(self):
        return get_models_module(self.get_folder_name()).Player

    @classmethod
    def _has_group_by_arrival_time(cls):
        app_name = cls.get_folder_name()
        return has_group_by_arrival_time(app_name)

    def group_randomly(self, *, fixed_id_in_group=False):
        group_matrix = self.get_group_matrix()
        group_matrix = otree.common._group_randomly(group_matrix, fixed_id_in_group)
        self.set_group_matrix(group_matrix)

    def creating_session(self):
        pass

    def vars_for_admin_report(self):
        return {}

    def _gbat_try_to_make_new_group(self, page_index):
        '''Returns the group ID of the participants who were regrouped'''
        from otree.models import Participant

        Player = self._PlayerClass()
        STALE_THRESHOLD_SECONDS = 70

        # count how many are re-grouped
        waiting_players = list(
            self.player_set.join(Participant).filter(
                Participant._waitpage_is_connected == True,
                Participant._waitpage_tab_hidden == False,
                Participant._index_in_pages == page_index,
                Participant._gbat_grouped == False,
                # this is just a failsafe
                Participant._last_request_timestamp
                >= time.time() - STALE_THRESHOLD_SECONDS,
            )
        )

        target = self.get_user_defined_target()
        # user may not have defined it
        func = getattr(
            target,
            'group_by_arrival_time_method',
            type(self).group_by_arrival_time_method,
        )
        players_for_group = func(self, waiting_players)

        if not players_for_group:
            return None

        participants = [p.participant for p in players_for_group]

        group_id_in_subsession = self._gbat_next_group_id_in_subsession()

        Constants = self._Constants
        num_rounds = Constants.get_normalized('num_rounds')
        this_round_new_group = None
        for round_number in range(self.round_number, num_rounds + 1):
            subsession = self.in_round(round_number)

            unordered_players = subsession.player_set.filter(
                Player.participant_id.in_([pp.id for pp in participants])
            )

            participant_ids_to_players = {
                player.participant: player for player in unordered_players
            }

            ordered_players_for_group = [
                participant_ids_to_players[participant] for participant in participants
            ]

            group = self._GroupClass()(
                subsession=subsession,
                id_in_subsession=group_id_in_subsession,
                session=self.session,
                round_number=round_number,
            )
            db.add(group)
            group.set_players(ordered_players_for_group)

            if round_number == self.round_number:
                this_round_new_group = group

            # note-to-self
            # prune groups without players
            # https://stackoverflow.com/a/21115972/
            for group_to_delete in subsession.group_set.outerjoin(Player).filter(
                Player.id == None
            ):
                db.delete(group_to_delete)

        for participant in participants:
            participant._gbat_grouped = True
            participant._waitpage_is_connected = False

        return this_round_new_group

    def _gbat_next_group_id_in_subsession(self):
        # 2017-05-05: seems like this can result in id_in_subsession that
        # doesn't start from 1.
        # especially if you do group_by_arrival_time in every round
        # is that a problem?
        Group = self._GroupClass()
        return (
            dbq(func.max(Group.id_in_subsession))
            .filter_by(session=self.session)
            .scalar()
            + 1
        )

    def group_by_arrival_time_method(self, waiting_players):
        Constants = self._Constants
        ppg = Constants.get_normalized('players_per_group')

        if ppg is None:
            msg = (
                'If using group_by_arrival_time, you must either set '
                'Constants.players_per_group to a value other than None, '
                'or define group_by_arrival_time_method.'
            )
            raise AssertionError(msg)

        if len(waiting_players) >= ppg:
            return waiting_players[:ppg]

    @declared_attr
    def group_set(cls):
        return relationship(
            f'{cls.__module__}.Group', back_populates="subsession", lazy='dynamic'
        )

    @declared_attr
    def player_set(cls):
        return relationship(
            f'{cls.__module__}.Player', back_populates="subsession", lazy='dynamic'
        )



import logging
from otree.database import session_scope

from otree.bots.runner import run_all_bots_for_session_config
from otree.constants import AUTO_NAME_BOTS_EXPORT_FOLDER
from .base import BaseCommand

logger = logging.getLogger('otree')
MSG_BOTS_HELP = 'Run oTree bots'


class Command(BaseCommand):
    help = MSG_BOTS_HELP

    def add_arguments(self, parser):
        # Positional arguments
        parser.add_argument(
            'session_config_name',
            nargs='?',
            help='If omitted, all sessions in SESSION_CONFIGS are run',
        )

        parser.add_argument(
            'num_participants',
            type=int,
            nargs='?',
            help='Number of participants (if omitted, use num_demo_participants)',
        )

        # don't call it --data because then people might think that
        # that's the *input* data folder
        parser.add_argument(
            '--export',
            nargs='?',
            const=AUTO_NAME_BOTS_EXPORT_FOLDER,
            dest='export_path',
            help=(
                'Saves the data generated by the tests. '
                'Runs the "export data" command, '
                'outputting the CSV files to the specified directory, '
                'or an auto-generated one.'
            ),
        )
        parser.add_argument(
            '--save',
            nargs='?',
            const=AUTO_NAME_BOTS_EXPORT_FOLDER,
            dest='export_path',
            help=('Alias for --export.'),
        )

    def handle(self, *, session_config_name, num_participants, export_path, **options):

        run_all_bots_for_session_config(
            session_config_name=session_config_name,
            num_participants=num_participants,
            export_path=export_path,
        )



import json
from pathlib import Path
from importlib import import_module

from starlette.requests import Request
from starlette.responses import Response, JSONResponse

import otree
import otree.bots.browser
import otree.views.cbv
from otree import settings
from otree.channels import utils as channel_utils
from otree.common import GlobalState, get_models_module
from otree.common2 import json_dumps
from otree.database import db, dbq
from otree.models import Session, Participant
from otree.models_concrete import ParticipantVarsFromREST
from otree.room import ROOM_DICT
from otree.session import create_session, SESSION_CONFIGS_DICT, CreateSessionInvalidArgs
from otree.templating import ibis_loader

from .cbv import BaseRESTView


class RESTOTreeVersion(BaseRESTView):
    url_pattern = '/api/otree_version'

    def get(self):
        return JSONResponse(dict(version=otree.__version__))


class RESTSessionConfigs(BaseRESTView):
    url_pattern = '/api/session_configs'

    def get(self):
        return Response(json_dumps(list(SESSION_CONFIGS_DICT.values())))


class RESTRooms(BaseRESTView):
    url_pattern = '/api/rooms'

    def get(self):
        data = [r.rest_api_dict(self.request) for r in ROOM_DICT.values()]
        return JSONResponse(data)


class RESTSessionVars(BaseRESTView):

    url_pattern = '/api/session_vars/{code}'

    def post(self, vars):
        code = self.request.path_params['code']
        session = db.get_or_404(Session, code=code)
        session.vars.update(vars)
        return JSONResponse({})


class RESTParticipantVars(BaseRESTView):

    url_pattern = '/api/participant_vars/{code}'

    def post(self, vars):
        code = self.request.path_params['code']
        participant = db.get_or_404(Participant, code=code)
        participant.vars.update(vars)
        return JSONResponse({})


class RESTParticipantVarsByRoom(BaseRESTView):
    """
    This can be used when you don't know the participant code,
    or when the participant doesn't have a code yet.
    For example, you might need to send data to oTree about the participant
    BEFORE sending the participant to oTree via their room link.
    """

    url_pattern = '/api/participant_vars'

    def post(self, room_name, participant_label, vars):
        if room_name not in ROOM_DICT:
            return Response(f'Room {room_name} not found', status_code=404)
        room = ROOM_DICT[room_name]
        session = room.get_session()
        if session:
            participant = session.pp_set.filter_by(label=participant_label).first()
            if participant:
                participant.vars.update(vars)
                return JSONResponse({})
        kwargs = dict(
            participant_label=participant_label,
            room_name=room_name,
        )
        _json_data = json.dumps(vars)
        obj = ParticipantVarsFromREST.objects_first(**kwargs)
        if obj:
            obj._json_data = _json_data
        else:
            obj = ParticipantVarsFromREST(**kwargs, _json_data=_json_data)
            db.add(obj)
        return JSONResponse({})


class RESTSessions(BaseRESTView):

    url_pattern = '/api/sessions'

    def get(self):
        sessions = []
        for session in dbq(Session).filter_by(is_demo=False).order_by('id'):
            session_dict = session_attrs_for_list(session)
            session_dict.update(get_session_urls(session, self.request))
            sessions.append(session_dict)
        return JSONResponse(sessions)

    def post(self, **kwargs):
        try:
            session = create_session(**kwargs)
        except CreateSessionInvalidArgs as exc:
            return Response(str(exc), status_code=400)
        room_name = kwargs.get('room_name')

        response_payload = dict(code=session.code)
        if room_name:
            channel_utils.sync_group_send(
                group=channel_utils.room_participants_group_name(room_name),
                data={'status': 'session_ready'},
            )

        response_payload.update(get_session_urls(session, self.request))

        return JSONResponse(response_payload)


def get_session_urls(session: Session, request: Request) -> dict:
    d = dict(
        session_wide_url=request.url_for(
            'SessionWideLink', anonymous_code=session._anonymous_code
        ),
        admin_url=request.url_for('SessionStartLinks', code=session.code),
    )
    room = session.get_room()
    if room:
        d['room_url'] = room.get_room_wide_url(request)
    return d


def session_attrs_for_list(session: Session):
    # this should include fields that help you distinguish it from other sessions

    return dict(
        code=session.code,
        num_participants=session.num_participants,
        created_at=session._created,
        label=session.label,
        config_name=session.config['name'],
    )


def session_attrs_for_detail(session: Session):
    return dict(
        # we need the session config for mturk settings and participation fee
        # technically, other parts of session config might not be JSON serializable
        config=session.config,
        REAL_WORLD_CURRENCY_CODE=settings.REAL_WORLD_CURRENCY_CODE,
    )


class RESTGetSessionInfo(BaseRESTView):
    # it's more of GET semantics but we need to use POST because we pass request.body
    url_pattern = '/api/get_session/{code}'

    def post(self, participant_labels=None, participant_vars=None, session_vars=None):
        if participant_vars is None:
            participant_vars = []
        if session_vars is None:
            session_vars = []
        code = self.request.path_params['code']
        session = db.get_or_404(Session, code=code)
        pp_set = session.pp_set
        if participant_labels is not None:
            pp_set = pp_set.filter(Participant.label.in_(participant_labels))
        pdata_list = []
        for pp in pp_set:
            pdata = dict(
                id_in_session=pp.id_in_session,
                code=pp.code,
                label=pp.label,
                payoff_in_real_world_currency=pp.payoff.to_real_world_currency(session),
                status=pp.status,
            )
            # shouldn't raise an error if the field is not found,
            # e.g. because oTree Hub may check if some things are defined to provide
            # extra functionality.
            # there is no silent failure because we will just not include that key
            # in the output, so it will be obvious.
            for field in participant_vars:
                if field in pp.vars:
                    val = pp.vars[field]
                    try:
                        json_dumps(val)
                    except TypeError:
                        return Response(
                            f"participant.vars['{field}'] is not JSON serializable",
                            status_code=400,
                        )
                    pdata[field] = val

            pdata_list.append(pdata)

        payload = session_attrs_for_list(session)
        payload.update(session_attrs_for_detail(session))
        payload.update(get_session_urls(session, self.request))

        session_config_filtered = {}
        for k in session.config:
            try:
                json_dumps(session.config[k])
            except TypeError:
                pass
            else:
                session_config_filtered[k] = session.config[k]

        payload.update(
            config=session_config_filtered,
            participants=pdata_list,
            REAL_WORLD_CURRENCY_CODE=settings.REAL_WORLD_CURRENCY_CODE,
        )

        for field in session_vars:
            if field in session.vars:
                val = session.vars[field]
                try:
                    json_dumps(val)
                except TypeError:
                    return Response(
                        f"session.vars['{field}'] is not JSON serializable",
                        status_code=400,
                    )
                payload[field] = val

        mturk_settings = session.config.get('mturk_hit_settings')
        if mturk_settings:
            payload['mturk_template_html'] = ibis_loader.search_template(
                mturk_settings['template']
            ).read_text('utf8')

        # need custom json_dumps for currency values
        return Response(json_dumps(payload))


class RESTGetSessionInfoLegacy(BaseRESTView):
    # for compat.
    # it used to be GET, but since it uses request.body changed to POST.
    url_pattern = '/api/sessions/{code}'
    get = RESTGetSessionInfo.post


launcher_session_code = None


class CreateBrowserBotsSession(BaseRESTView):
    url_pattern = '/create_browser_bots_session'

    def post(
        self,
        num_participants,
        session_config_name,
        case_number,
    ):
        session = create_session(
            session_config_name=session_config_name, num_participants=num_participants
        )
        otree.bots.browser.initialize_session(
            session_pk=session.id, case_number=case_number
        )
        GlobalState.browser_bots_launcher_session_code = session.code
        channel_utils.sync_group_send(
            group='browser_bot_wait', data={'status': 'session_ready'}
        )

        return Response(session.code)


class CloseBrowserBotsSession(BaseRESTView):
    url_pattern = '/close_browser_bots_session'

    def post(self, **kwargs):
        GlobalState.browser_bots_launcher_session_code = None
        return Response('ok')


class RESTApps(BaseRESTView):
    url_pattern = '/api/apps'

    def get(self):
        from otree.settings import OTREE_APPS

        d = {}
        for app in OTREE_APPS:
            models_module = get_models_module(app)
            d[app] = getattr(models_module, 'doc', '')
        return Response(json_dumps(d))



import os

from otree import settings
from otree.session import SESSION_CONFIGS_DICT
from .cbv import AdminView


class DemoIndex(AdminView):
    url_pattern = '/demo'

    def vars_for_template(self):
        title = getattr(settings, 'DEMO_PAGE_TITLE', 'Demo')
        intro_html = getattr(settings, 'DEMO_PAGE_INTRO_HTML', '')
        session_info = []
        for session_config in SESSION_CONFIGS_DICT.values():
            session_info.append(
                {
                    'name': session_config['name'],
                    'display_name': session_config['display_name'],
                    'url': self.request.url_for(
                        'CreateDemoSession', config_name=session_config['name']
                    ),
                    'num_demo_participants': session_config['num_demo_participants'],
                }
            )

        if os.environ.get('OTREEHUB_PUB'):
            otreehub_app_name = os.environ.get('OTREEHUB_APP_NAME')
            otreehub_url = f'https://www.otreehub.com/projects/{otreehub_app_name}/'
        else:
            otreehub_url = ''

        return dict(
            session_info=session_info,
            title=title,
            intro_html=intro_html,
            is_debug=settings.DEBUG,
            otreehub_url=otreehub_url,
        )


class CreateDemoSession(AdminView):
    url_pattern = '/demo/{config_name}'

    def vars_for_template(self):
        return dict(config_name=self.request.path_params['config_name'])



import asyncio

import starlette.exceptions
from starlette.concurrency import run_in_threadpool
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.types import Message, Receive, Scope, Send


class ExceptionMiddleware(starlette.exceptions.ExceptionMiddleware):
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """oTree just removed the 'from None'. everything else is the same
        Need this until https://github.com/encode/starlette/issues/1114 is fixed
        """
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        response_started = False

        async def sender(message: Message) -> None:
            nonlocal response_started

            if message["type"] == "http.response.start":
                response_started = True
            await send(message)

        try:
            await self.app(scope, receive, sender)
        except Exception as exc:
            handler = None

            if isinstance(exc, HTTPException):
                handler = self._status_handlers.get(exc.status_code)

            if handler is None:
                handler = self._lookup_exception_handler(exc)

            if handler is None:
                # oTree changed this part only
                raise exc  # from None

            if response_started:
                msg = "Caught handled exception, but response already started."
                raise RuntimeError(msg) from exc

            request = Request(scope, receive=receive)
            if asyncio.iscoroutinefunction(handler):
                response = await handler(request, exc)
            else:
                response = await run_in_threadpool(handler, request, exc)
            await response(scope, receive, sender)



import contextlib
import json
import logging
from collections import defaultdict, namedtuple
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from typing import List, Dict, Union, Optional

from starlette.requests import Request
from starlette.responses import Response

from otree import settings
from otree.database import values_flat, db
from otree.models import Session, Participant
from otree.mturk_client import TurkClient, MTurkError
from otree.templating import ibis_loader
from otree.views.cbv import AdminSessionPage
from .cbv import enqueue_admin_message

logger = logging.getLogger('otree')


@dataclass
class MTurkSettings:
    keywords: Union[str, list]
    title: str
    description: str
    frame_height: int
    template: str
    minutes_allotted_per_assignment: int
    expiration_hours: float
    qualification_requirements: List
    grant_qualification_id: Optional[str] = None


@contextlib.contextmanager
def ReportMTurkErrors():
    '''Alternative to get_mturk_client, for when we need exception handling
    in admin views, we should pass it, so that we can show the user the message
    without crashing.
    for participant-facing views and commandline tools, should use get_mturk_client.
    '''
    try:
        yield
    except MTurkError as exc:
        logger.error('MTurk error', exc_info=True)
        enqueue_admin_message('danger', repr(exc))


def in_public_domain(request: Request):
    """This method validates if oTree are published on a public domain
    because mturk need it

    """
    host = request.url.hostname.lower()
    if ":" in host:
        host = host.split(":", 1)[0]
    if host in ["localhost", '127.0.0.1']:
        return False
    # IPy had a compat problem with py 3.8.
    # in the future, could move some IPy code here.
    return True


class MTurkCreateHIT(AdminSessionPage):

    # make these class attributes so they can be mocked
    aws_keys_exist = bool(
        getattr(settings, 'AWS_ACCESS_KEY_ID', None)
        and getattr(settings, 'AWS_SECRET_ACCESS_KEY', None)
    )

    def vars_for_template(self):
        session = self.session

        mturk_settings = session.config['mturk_hit_settings']

        is_usd = settings.REAL_WORLD_CURRENCY_CODE == 'USD'
        mturk_ready = self.aws_keys_exist and is_usd

        return dict(
            mturk_settings=mturk_settings,
            participation_fee=session.config['participation_fee'],
            mturk_num_workers=session.mturk_num_workers(),
            mturk_ready=mturk_ready,
            aws_keys_exist=self.aws_keys_exist,
            is_usd=is_usd,
        )

    def post(self, request, code):
        session = self.session
        use_sandbox = bool(self.get_post_data().get('use_sandbox'))
        if not in_public_domain(request) and not use_sandbox:
            msg = 'oTree must run on a public domain for Mechanical Turk (e.g. not a localhost server)'
            return Response(msg)
        mturk_settings = MTurkSettings(**session.config['mturk_hit_settings'])

        start_url = self.request.url_for('MTurkStart', code=session.code)

        keywords = mturk_settings.keywords
        if isinstance(keywords, (list, tuple)):
            keywords = ', '.join(keywords)

        html_question = ibis_loader.load('otree/MTurkHTMLQuestion.html').render(
            user_template=mturk_settings.template,
            frame_height=mturk_settings.frame_height,
            start_url=start_url,
        )

        mturk_hit_parameters = {
            'Title': mturk_settings.title,
            'Description': mturk_settings.description,
            'Keywords': keywords,
            'MaxAssignments': session.mturk_num_workers(),
            'Reward': str(float(session.config['participation_fee'])),
            'AssignmentDurationInSeconds': 60
            * mturk_settings.minutes_allotted_per_assignment,
            'LifetimeInSeconds': int(60 * 60 * mturk_settings.expiration_hours),
            # prevent duplicate HITs
            'UniqueRequestToken': 'otree_{}'.format(session.code),
            'Question': html_question,
        }

        if not use_sandbox:
            # drop requirements checks in sandbox mode.
            mturk_hit_parameters['QualificationRequirements'] = (
                mturk_settings.qualification_requirements
            )

        with ReportMTurkErrors():

            hit = TurkClient.create_hit(mturk_hit_parameters, use_sandbox=use_sandbox)[
                'HIT'
            ]

            session.mturk_HITId = hit['HITId']
            session.mturk_HITGroupId = hit['HITGroupId']
            session.mturk_use_sandbox = use_sandbox
            session.mturk_expiration = hit['Expiration']
            session.mturk_qual_id = mturk_settings.grant_qualification_id or ''

        return self.redirect('MTurkCreateHIT', code=session.code)


Assignment = namedtuple(
    'Assignment', ['worker_id', 'assignment_id', 'status', 'answer']
)


def get_all_assignments(hit_id, *, use_sandbox) -> List[Assignment]:
    # Accumulate all relevant assignments, one page of results at
    # a time.
    assignments = []

    args = dict(
        HITId=hit_id,
        # i think 100 is the max page size
        MaxResults=100,
        AssignmentStatuses=['Submitted', 'Approved', 'Rejected'],
    )

    while True:
        response = TurkClient.list_assignments_for_hit(args, use_sandbox=use_sandbox)
        if not response['Assignments']:
            break
        for d in response['Assignments']:
            assignments.append(
                Assignment(
                    worker_id=d['WorkerId'],
                    assignment_id=d['AssignmentId'],
                    status=d['AssignmentStatus'],
                    answer=d['Answer'],
                )
            )
        args['NextToken'] = response['NextToken']

    return assignments


def get_workers_by_status(
    all_assignments: List[Assignment],
) -> Dict[str, List[Assignment]]:
    workers_by_status = defaultdict(list)
    for assignment in all_assignments:
        workers_by_status[assignment.status].append(assignment.worker_id)
    return workers_by_status


class MTurkSessionPayments(AdminSessionPage):
    def vars_for_template(self):
        session = self.session
        if not session.mturk_HITId:
            return dict(published=False)

        with ReportMTurkErrors():
            all_assignments = get_all_assignments(
                hit_id=session.mturk_HITId, use_sandbox=session.mturk_use_sandbox
            )

            # auto-reject logic
            assignment_ids_in_db = values_flat(
                session.pp_set.filter(Participant.mturk_assignment_id != None),
                'mturk_assignment_id',
            )

            submitted_assignment_ids = [
                a.assignment_id for a in all_assignments if a.status == 'Submitted'
            ]

            auto_rejects = set(submitted_assignment_ids) - set(assignment_ids_in_db)

            for assignment_id in auto_rejects:
                TurkClient.reject_assignment(
                    dict(
                        AssignmentId=assignment_id,
                        RequesterFeedback='Auto-rejecting because this assignment was not found in our database.',
                    ),
                    use_sandbox=session.mturk_use_sandbox,
                )

        workers_by_status = get_workers_by_status(all_assignments)

        def get_participants_by_status(status):
            return list(
                session.pp_set.filter(
                    Participant.mturk_worker_id.in_(workers_by_status[status])
                )
            )

        participants_approved = get_participants_by_status('Approved')
        participants_rejected = get_participants_by_status('Rejected')
        participants_not_reviewed = get_participants_by_status('Submitted')

        for lst in [
            participants_not_reviewed,
            participants_approved,
            participants_rejected,
        ]:
            add_answers(lst, all_assignments)

        return dict(
            published=True,
            participants_approved=participants_approved,
            participants_rejected=participants_rejected,
            participants_not_reviewed=participants_not_reviewed,
            participation_fee=session.config['participation_fee'],
            auto_rejects=auto_rejects,
        )


def get_completion_code(xml: str) -> str:
    if not xml:
        return ''
    # move inside function because it adds 0.03s to startup time
    from xml.etree import ElementTree

    root = ElementTree.fromstring(xml)
    for ans in root:
        if ans[0].text == 'taskAnswers':
            answer_data = json.loads(ans[1].text)
            try:
                return answer_data[0]['completion_code']
            except:
                return ''
    return ''


def add_answers(participants: List[Participant], all_assignments: List[Assignment]):
    answers = {}
    for assignment in all_assignments:
        answers[assignment.worker_id] = assignment.answer
    for p in participants:
        p._is_frozen = False
        p.mturk_answers_formatted = get_completion_code(answers[p.mturk_worker_id])


class PayMTurk(AdminSessionPage):
    """only POST"""

    url_pattern = '/PayMTurk/{code}'

    def post(self, request, code):
        session = db.get_or_404(Session, code=code)
        successful_payments = 0
        failed_payments = 0
        post_data = self.get_post_data()

        payment_page_response = self.redirect('MTurkSessionPayments', code=session.code)
        # use worker ID instead of assignment ID. Because 2 workers can have
        # the same assignment (if 1 starts it then returns it). we can't really
        # block that.
        # however, we can ensure that 1 worker does not get 2 assignments,
        # by enforcing that the same worker is always assigned to the same participant.
        participants = session.pp_set.filter(
            Participant.mturk_worker_id.in_(post_data.getlist('workers'))
        )

        for p in participants:
            payoff = p.payoff.to_real_world_currency(session)

            # need the try/except so that we try to pay the rest of the participants
            try:
                if payoff > 0:
                    TurkClient.send_bonus(
                        dict(
                            WorkerId=p.mturk_worker_id,
                            AssignmentId=p.mturk_assignment_id,
                            BonusAmount='{0:.2f}'.format(Decimal(payoff)),
                            # prevent duplicate payments
                            UniqueRequestToken='{}_{}'.format(
                                p.mturk_worker_id, p.mturk_assignment_id
                            ),
                            # this field is required.
                            Reason='Thank you',
                        ),
                        use_sandbox=session.mturk_use_sandbox,
                    )

                # approve assignment should happen AFTER bonus, so that if bonus fails,
                # the user will still show up in assignments_not_reviewed.
                # worst case is that bonus succeeds but approval fails.
                # in that case, exception will be raised on send_bonus because of UniqueRequestToken.
                # but that's OK, then you can just unselect that participant and pay the others.
                TurkClient.approve_assignment(
                    dict(AssignmentId=p.mturk_assignment_id),
                    use_sandbox=session.mturk_use_sandbox,
                )
                successful_payments += 1
            except Exception as e:
                msg = (
                    'Could not pay {} because of an error communicating '
                    'with MTurk: {}'.format(p._numeric_label(), str(e))
                )
                enqueue_admin_message('danger', msg)
                logger.error(msg)
                failed_payments += 1
                if failed_payments > 10:
                    return payment_page_response
        msg = 'Successfully made {} payments.'.format(successful_payments)
        if failed_payments > 0:
            msg += ' {} payments failed.'.format(failed_payments)
            enqueue_admin_message('warning', msg)
        else:
            enqueue_admin_message('success', msg)
        return payment_page_response


class RejectMTurk(AdminSessionPage):
    """POST only"""

    url_pattern = '/RejectMTurk/{code}'

    def post(self, request, code):
        session = db.get_or_404(Session, code=code)
        with ReportMTurkErrors():
            for p in session.pp_set.filter(
                Participant.mturk_worker_id.in_(self.get_post_data().getlist('workers'))
            ):
                TurkClient.reject_assignment(
                    dict(
                        AssignmentId=p.mturk_assignment_id,
                        # The boto3 docs say this param is optional, but if I omit it, I get:
                        # An error occurred (ValidationException) when calling the RejectAssignment operation:
                        # 1 validation error detected: Value null at 'requesterFeedback'
                        # failed to satisfy constraint: Member must not be null
                        RequesterFeedback='',
                    ),
                    use_sandbox=session.mturk_use_sandbox,
                )

            enqueue_admin_message('success', "Rejected the selected assignments")
        return self.redirect('MTurkSessionPayments', code=code)


class MTurkExpireHIT(AdminSessionPage):
    """only POST"""

    url_pattern = '/MTurkExpireHIT/{code}'

    def post(self, request, code):
        session = db.get_or_404(Session, code=code)
        with ReportMTurkErrors():
            expiration = datetime(2015, 1, 1).timestamp()
            TurkClient.update_expiration(
                dict(
                    HITId=session.mturk_HITId,
                    # If you update it to a time in the past,
                    # the HIT will be immediately expired.
                    ExpireAt=expiration,
                ),
                use_sandbox=session.mturk_use_sandbox,
            )
            session.mturk_expiration = expiration
        # don't need a message because the MTurkCreateHIT page will
        # statically say the HIT has expired.
        return self.redirect('MTurkCreateHIT', code=code)



import datetime
import hashlib
import hmac
import json
from datetime import datetime
from urllib.error import HTTPError
from urllib.request import urlopen, Request

from otree.settings import AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY


# Key derivation functions. See:
# http://docs.aws.amazon.com/general/latest/gr/signature-v4-examples.html#signature-v4-examples-python
def sign(key, msg):
    return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()


def getSignatureKey(key, date_stamp, regionName, serviceName):
    kDate = sign(('AWS4' + key).encode('utf-8'), date_stamp)
    kRegion = sign(kDate, regionName)
    kService = sign(kRegion, serviceName)
    kSigning = sign(kService, 'aws4_request')
    return kSigning


def call_api(operation: str, request_parameters: dict, *, use_sandbox: bool) -> dict:
    # ************* REQUEST VALUES *************

    if use_sandbox:
        host = 'mturk-requester-sandbox.us-east-1.amazonaws.com'
    else:
        host = 'mturk-requester.us-east-1.amazonaws.com'
    endpoint = f'https://{host}'

    service = 'mturk-requester'
    region = 'us-east-1'
    # POST requests use a content type header. For DynamoDB,
    # the content is JSON.
    content_type = 'application/x-amz-json-1.1'
    amz_target = f'MTurkRequesterServiceV20170117.{operation}'

    # Create a date for headers and the credential string
    t = datetime.utcnow()
    amz_date = t.strftime('%Y%m%dT%H%M%SZ')
    date_stamp = t.strftime('%Y%m%d')  # Date w/o time, used in credential scope

    # ************* TASK 1: CREATE A CANONICAL REQUEST *************
    # http://docs.aws.amazon.com/general/latest/gr/sigv4-create-canonical-request.html

    # Step 2: Create canonical URI--the part of the URI from domain to query
    # string (use '/' if no path)
    canonical_uri = '/'

    ## Step 3: Create the canonical query string. In this example, request
    # parameters are passed in the body of the request and the query string
    # is blank.
    canonical_querystring = ''

    # Step 4: Create the canonical headers. Header names must be trimmed
    # and lowercase, and sorted in code point order from low to high.
    # Note that there is a trailing \n.
    canonical_headers = (
        'content-type:'
        + content_type
        + '\n'
        + 'host:'
        + host
        + '\n'
        + 'x-amz-date:'
        + amz_date
        + '\n'
        + 'x-amz-target:'
        + amz_target
        + '\n'
    )

    # Step 5: Create the list of signed headers. This lists the headers
    # in the canonical_headers list, delimited with ";" and in alpha order.
    # Note: The request can include any headers; canonical_headers and
    # signed_headers include those that you want to be included in the
    # hash of the request. "Host" and "x-amz-date" are always required.
    # For DynamoDB, content-type and x-amz-target are also required.
    signed_headers = 'content-type;host;x-amz-date;x-amz-target'

    request_body = json.dumps(request_parameters).encode('utf8')
    # Step 6: Create payload hash. In this example, the payload (body of
    # the request) contains the request parameters.
    payload_hash = hashlib.sha256(request_body).hexdigest()

    # Step 7: Combine elements to create canonical request
    canonical_request = '\n'.join(
        [
            'POST',
            canonical_uri,
            canonical_querystring,
            canonical_headers,
            signed_headers,
            payload_hash,
        ]
    )

    # ************* TASK 2: CREATE THE STRING TO SIGN*************
    # Match the algorithm to the hashing algorithm you use, either SHA-1 or
    # SHA-256 (recommended)
    algorithm = 'AWS4-HMAC-SHA256'

    credential_scope = '/'.join([date_stamp, region, service, 'aws4_request'])
    string_to_sign = '\n'.join(
        [
            algorithm,
            amz_date,
            credential_scope,
            hashlib.sha256(canonical_request.encode('utf-8')).hexdigest(),
        ]
    )

    # ************* TASK 3: CALCULATE THE SIGNATURE *************
    # Create the signing key using the function defined above.
    signing_key = getSignatureKey(AWS_SECRET_ACCESS_KEY, date_stamp, region, service)

    # Sign the string_to_sign using the signing_key
    signature = hmac.new(
        signing_key, string_to_sign.encode('utf-8'), hashlib.sha256
    ).hexdigest()

    # ************* TASK 4: ADD SIGNING INFORMATION TO THE REQUEST *************
    # Put the signature information in a header named Authorization.

    authorization_header = f"{algorithm} Credential={AWS_ACCESS_KEY_ID}/{credential_scope}, SignedHeaders={signed_headers}, Signature={signature}"

    # For DynamoDB, the request can include any headers, but MUST include "host", "x-amz-date",
    # "x-amz-target", "content-type", and "Authorization". Except for the authorization
    # header, the headers must be included in the canonical_headers and signed_headers values, as
    # # Python note: The 'host' header is added automatically by the Python 'requests' library.
    headers = {
        'Content-Type': content_type,
        'X-Amz-Date': amz_date,
        'X-Amz-Target': amz_target,
        'Authorization': authorization_header,
    }

    request = Request(endpoint, headers=headers)
    try:
        resp = urlopen(request, data=request_body)
    except HTTPError as exc:
        # 2022-12-05: I used to use ['Message'],
        # but it seems this key is not always present.
        msg = json.loads(exc.read().decode('utf8'))
        raise MTurkError(msg)
    res_body = resp.read()
    return json.loads(res_body.decode("utf-8"))


class MTurkError(Exception):
    pass


class TurkClient:
    """
    - Make these separate methods so they are easier to mock individually
    - Put them under a class so that they are easier to patch (don't have to worry about
    how it's imported)

    Also, in the future we might add some state, e.g. persistent connection
    """

    @staticmethod
    def create_hit(request_params, *, use_sandbox):
        return call_api('CreateHIT', request_params, use_sandbox=use_sandbox)

    @staticmethod
    def list_assignments_for_hit(request_params, *, use_sandbox):
        return call_api(
            'ListAssignmentsForHIT', request_params, use_sandbox=use_sandbox
        )

    @staticmethod
    def reject_assignment(request_params, *, use_sandbox):
        return call_api('RejectAssignment', request_params, use_sandbox=use_sandbox)

    @staticmethod
    def approve_assignment(request_params, *, use_sandbox):
        return call_api('ApproveAssignment', request_params, use_sandbox=use_sandbox)

    @staticmethod
    def send_bonus(request_params, *, use_sandbox):
        return call_api('SendBonus', request_params, use_sandbox=use_sandbox)

    @staticmethod
    def update_expiration(request_params, *, use_sandbox):
        return call_api(
            'UpdateExpirationForHIT', request_params, use_sandbox=use_sandbox
        )

    @staticmethod
    def assign_qualification(request_params, *, use_sandbox):
        return call_api(
            'AssociateQualificationWithWorker', request_params, use_sandbox=use_sandbox
        )



from sqlalchemy import Column, ForeignKey
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship
from sqlalchemy.sql import sqltypes as st

from otree.common import in_round, in_rounds
from otree.database import db, MixinSessionFK, SPGModel, CurrencyType


class BasePlayer(SPGModel, MixinSessionFK):
    __abstract__ = True

    id_in_group = Column(st.Integer, nullable=True, index=True,)

    # don't modify this directly! Set player.payoff instead
    _payoff = Column(CurrencyType, default=0)

    round_number = Column(st.Integer, index=True)

    # make it non-nullable so that we don't raise an error with null.
    # the reason i chose to make this different from ordinary StringFields
    # is that it's a property. users can't just use .get('role') because
    # that will just access ._role. So we would need some special-casing
    # in __getattribute__ for role, which is not desirable.
    _role = Column(st.String, nullable=False, default='')

    # as a property, that means it's overridable
    @property
    def role(self):
        return self._role

    @property
    def payoff(self):
        return self._payoff

    @payoff.setter
    def payoff(self, value):
        if value is None:
            value = 0
        delta = value - self._payoff
        self._payoff += delta
        self.participant.payoff += delta
        # should save it because it may not be obvious that modifying
        # player.payoff also changes a field on a different model
        db.commit()

    @property
    def id_in_subsession(self):
        return self.participant.id_in_session

    def in_round(self, round_number):
        return in_round(type(self), round_number, participant=self.participant)

    def in_rounds(self, first, last):
        return in_rounds(type(self), first, last, participant=self.participant)

    def in_previous_rounds(self):
        return self.in_rounds(1, self.round_number - 1)

    def in_all_rounds(self):
        '''i do it this way because it doesn't rely on idmap'''
        return self.in_previous_rounds() + [self]

    def get_others_in_group(self):
        return [p for p in self.group.get_players() if p != self]

    def get_others_in_subsession(self):
        return [p for p in self.subsession.get_players() if p != self]

    def start(self):
        pass

    @declared_attr
    def subsession_id(cls):
        app_name = cls.get_folder_name()
        return Column(
            st.Integer, ForeignKey(f'{app_name}_subsession.id', ondelete='CASCADE')
        )

    @declared_attr
    def subsession(cls):
        return relationship(f'{cls.__module__}.Subsession', back_populates='player_set')

    @declared_attr
    def group_id(cls):
        app_name = cls.get_folder_name()
        # needs to be nullable so re-grouping can happen
        return Column(st.Integer, ForeignKey(f'{app_name}_group.id'), nullable=True)

    @declared_attr
    def group(cls):
        return relationship(f'{cls.__module__}.Group', back_populates='player_set')

    @declared_attr
    def participant_id(cls):
        return Column(st.Integer, ForeignKey('otree_participant.id'))

    @declared_attr
    def participant(cls):
        return relationship("Participant")
